#include "cuddInt.h"
#include "message.h"

/* note theese function does not derefrence the nodes in f or g */

DdNodePar cuddBddIteBase(manager,f,g,h,op)
DdManager *manager;
DdNodePar  f;
DdNodePar  g;
DdNodePar  h;
int *op;
{
  DdNodePar one, zero;
  DdNodePar not_g,not_h;
  one = DD_ONE_PAR(manager);
  zero = Cudd_NotPar(one);
  
  /* One variable cases. */
  if (cuddEqualPar(f,one)) 	/* ITE(1,G,H) = G */
    return(g);
    
  if (cuddEqualPar(f,zero)) 	/* ITE(0,G,H) = H */
    return(h);
    
  /* ITE(F,F,H) = ITE(F,1,H) = F + H */
  if (cuddEqualPar(g,one) || cuddEqualPar(f,g)) { 
    *op=CUDD_BDD_OR_1_3;
    return(Cudd_NotPar(cuddBddAndBasePar(manager,Cudd_NotPar(f),Cudd_NotPar(g))));
  }

  not_h=Cudd_NotPar(h);
  /* ITE(F,G,!F) = ITE(F,G,1) = !F + G */
  if (cuddEqualPar(h,one) || cuddEqualPar(f,not_h)) { 
    *op=CUDD_BDD_OR_NOT1_2;
    return(Cudd_NotPar(cuddBddAndBasePar(manager,f,Cudd_NotPar(g))));
  }
  
  not_g=Cudd_NotPar(g);
  /* ITE(F,!F,H) = ITE(F,0,H) = !F * H */
  if (cuddEqualPar(g,zero) || cuddEqualPar(f,not_g)) { 
    *op=CUDD_BDD_AND_NOT1_3;
    return(cuddBddAndBasePar(manager,Cudd_NotPar(f),h));
  }

  /* ITE(F,G,F) = ITE(F,G,0) = F * G */
  if (cuddEqualPar(h,zero) || cuddEqualPar(f,h)) {    
    *op=CUDD_BDD_AND_1_2;
    return(cuddBddAndBasePar(manager,f,g));
  }

  
  /* Check remaining one variable case. */
  if (cuddEqualPar(g,h)) {     /* ITE(F,G,G) = G */
    return(g);
  }
  *op=CUDD_BDD_ITE;
  return(CUDD_ERROR);
}

DdNodePar cuddBddXorBasePar(manager,f,g)
DdManager *manager;
DdNodePar  f;
DdNodePar  g;
{
  DdNodePar one, zero;
  one = DD_ONE_PAR(manager);
  zero = Cudd_NotPar(one);
  
  /* Terminal cases. */
  if (cuddEqualPar(f,g)) return(zero);
  if (cuddEqualPar(f,Cudd_NotPar(g))) return(one);
  if (cuddEqualPar(f,one)) return(Cudd_NotPar(g));
  if (cuddEqualPar(g,one)) return(Cudd_NotPar(f));
  if (cuddEqualPar(f,zero)) return(g);
  if (cuddEqualPar(g,zero)) return(f);
  
  return(CUDD_ERROR);
} /* this was taken direct from orig code => no else after returns */



/* note this function does not derefrence the nodes in f or g */
DdNodePar cuddBddAndBasePar(manager,f,g)
DdManager *manager;
DdNodePar  f;
DdNodePar  g;
{

  /* OPTIMIZED
  if(f.index==CUDD_MAXINDEX)
    if(f.address.number == -1) return(f); 
    else                       return(g);
  else if(g.index==CUDD_MAXINDEX)
    if(g.address.number == -1) return(g);
    else                       return(f);
  else if(f.address.number==-g.address.number 
	  && f.address.page==g.address.page 
	  && f.index==g.index)
    return(Cudd_NotPar(DD_ONE_PAR(manager)));
  else if(f.address.number==g.address.number 
	  && f.address.page==g.address.page 
	  && f.index==g.index)
    return(f);
  */
  DdNodePar one, zero;
  
  one = DD_ONE_PAR(manager);
  zero = Cudd_NotPar(one);

  if (cuddEqualPar(f,zero) 
      || cuddEqualPar(g,zero) 
      || cuddEqualPar(f,Cudd_NotPar(g)))
    return(zero);
  else if (cuddEqualPar(f,g) || cuddEqualPar(g,one)) return(f);
  else if (cuddEqualPar(f,one)) return(g);

  else return(CUDD_ERROR);
}   

