#include "cuddInt.h"
#include "message.h"

int ref_=0;
int tot_forwards=0;

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/
static void cuddProfile(DdManager *manager,int index1,int index2);
static int cuddFindProfile(DdManager *manager);
static int cuddInitReorder(DdManager *manager);
static int cuddEndReorder(DdManager *manager);
static int cuddInitMore(DdManager *manager);
static int cuddChangeSubtableTypes(DdManager *manager);
static int cuddInitMemForReconstruct(DdManager *manager,DdMemory **old_mems);
static int cuddCreateCentralFreeList(DdManager *manager);
static int cuddRestoreSubtableTypes(DdManager *manager);
static int cuddRestoreLocalFreeLists(DdManager *manager);
static int cuddFindNewLocations(DdManager *manager,DdMemory *old_mems,
				DdNodeAdr nodelists[]);
static int cuddRecreateSubtables(DdManager *manager,DdMemory *old_mems,
				 DdNodeAdr lists[]);
static int cuddFreeMore(DdManager *manager,DdMemory *old_mems);
static int cuddSwap(DdManager *manager, int level);
static void cuddSwapNodes(DdManager *manager, int x_index, int y_index,DdNodePar list);
static int cuddResetPerm(DdManager *manager);
static void cuddSiftUp(DdManager *manager, int index, int top, int *best_pos, int *best_size);
static void cuddSiftDown(DdManager *manager, int index, int bottom, int *best_pos, int *best_size);
static void cuddSiftBlock(DdManager *manager,int block,int first_level, int last_level);
static void cuddBlockSift(DdManager *manager);

/* external functions */
int cuddReorder(DdManager *manager)
{
  cuddInitReorder(manager);
  cuddFindProfile(manager);
  cuddSendProfile(manager);

  cuddReorderRedistributeLevelsSlave(manager);
  /* could do selective recomputation of references, but easier to redo all */
  cuddResetRefs(manager);
  cuddGetExternRef(manager);
  cuddFindRef(manager);
  cuddSendExternRef(manager);

  /* this might be done in init, but easier to find refs with old structure */
  cuddChangeSubtableTypes(manager);
  cuddCreateCentralFreeList(manager);

  cuddBlockSift(manager);
  cuddEndReorder(manager);
  free(manager->Profile+1);

  return(1); /* NO ERROR HAMDELING */
}

/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/

static void cuddProfile(DdManager *manager,int index1,int index2)
{
  int i;
  if(index2==CUDD_MAXINDEX) index2=manager->size;
  for(i=index2;i>index1;i--)
    {
      manager->Profile[i]++;
    }
}

static int cuddFindProfile(DdManager *manager)
{
  int index,page_number,i,max;
  DdNodeOld *page;
  DdMemory *mem;
  DdNodeMore *ref_page;
  DdSubtable *subtable;

  manager->Profile=ALLOC(int,manager->size)-1;
  for(i=1;i<=manager->size;i++)
    manager->Profile[i]=0;
  
  for(index=0;index<manager->size;index++)
    if(cuddLocal(manager,index))
      {
	mem=&manager->memory[index];
	subtable=&manager->subtables[index];

	for(page_number=0;page_number<mem->nextPage;page_number++)
	  { 
	    if(page_number==mem->nextPage-1)
	      { /* last page */
		if(cuddEqualAdr(mem->nextFree,NULL_ADR))
		  max=DD_MEM_CHUNK;
		else
		  max=mem->nextFree.number-1;
	      }
	    else
	      max=DD_MEM_CHUNK;
	    
	    page=mem->PageTable[page_number];
	    ref_page=mem->PageTable2[page_number];
	    for(i=1;i<=max;i++) 
	      {
		if(ref_page[i].Reorder.Ref>0) {
		  cuddProfile(manager,index,page[i].type.kids.T.index);
		  cuddProfile(manager,index,page[i].type.kids.E.index);
		} 
	      }
	  }
      }
  return(1);
}

static int cuddInitReorder(DdManager *manager)
{
  manager->reordering=TRUE;

  FREE(manager->cache); 
  manager->memused -= manager->cacheSlots * sizeof(DdCache);

  /* this free up space and clears cache when it is reinitialized */

  cuddInitMore(manager);

  cuddReorderUpdateTableSizesSlave(manager);
  cuddGetExternRef(manager);

  cuddFindRef(manager);
  cuddSendExternRef(manager);
  return(1); /* NO ERROR HAMDELING */
}

static int cuddEndReorder(DdManager *manager)
{
  DdMemory *old_mems;
  DdNodeAdr nodelists[manager->size];

  manager->reordering=FALSE;

  cuddInitMemForReconstruct(manager,&old_mems);

  cuddRestoreSubtableTypes(manager);

  cuddInitExternDistributeNewLocations(manager,old_mems);

  cuddFindNewLocations(manager,old_mems,nodelists);

  cuddExternGetNewLocations(manager,old_mems);
  /* need to get new before sending own, because some of new may have to
     be passed on */
  cuddExternDistributeNewLocations(manager,old_mems);

  cuddRecreateSubtables(manager,old_mems,nodelists);

  cuddFreeMore(manager,old_mems);
  
  cuddResetPerm(manager);

  cuddInitCachePar(manager,manager->cacheSlots,manager->maxCacheHard);

  return(1); /* NO ERROR HAMDELING */
}

static int cuddResetPerm(DdManager *manager)
{
  int i;
  for(i=0;i<manager->size;i++)
    {
      cuddSetNewLevel(manager,i,i);
    }
}

static int cuddInitMore(DdManager *manager)
{
  int page_number,i,index;
  DdNodeMore *more_page;
  DdMemory *mem;

  for(index=0;index<manager->maxSize;index++)
    if(cuddLocal(manager,index)) {

      mem = &manager->memory[index];
      
      mem->PageTable2 = ALLOC(DdNodeMore *,mem->TableSize);
      if(mem->PageTable2 == NULL)
	return(0);
      
      for(page_number=0;page_number<mem->nextPage;page_number++)
	{ 
	  
	  more_page=mem->PageTable2[page_number]=ALLOC(DdNodeMore,DD_MEM_CHUNK)-1;
	  if(more_page==NULL)
	    return(0);
	  
	  manager->memused+=(DD_MEM_CHUNK) * sizeof(DdNodeMore);
	  
	  for(i=1;i<=DD_MEM_CHUNK;i++) {
	    more_page[i].Reorder.Ref=0;
	    more_page[i].Reorder.Index=index;
	    more_page[i].Reorder.NextIndex=index;
	  }
	}
    }
  return(1);
}

static int cuddChangeSubtableTypes(DdManager *manager)
{
  int index,i;
  DdNodeAdr *nodelist;
  DdNodePar *newlist;
  
  for(index=0;index<manager->maxSize;index++)
    if(cuddLocal(manager,index)) {
      nodelist=manager->subtables[index].nodelist;
      manager->subtables[index].ReorderNodelist=newlist
	=ALLOC(DdNodePar,manager->subtables[index].slots);
      if(newlist==NULL)
	return(0);

      for(i=0;i<manager->subtables[index].slots;i++)
	{
	  newlist[i].address=nodelist[i];
	  newlist[i].index=index;
	}
      FREE(nodelist);
    }
  return(1);
}

/* note central free list is strung through the then child rather
   then next pointer as 3-value addresses are needed */
static int cuddCreateCentralFreeList(DdManager *manager)
{
  int index;
  DdNodeOld *f_ptr;
  DdNodePar f,central_list=CUDD_ERROR;

  for(index=0;index<manager->size;index++)
    if(cuddLocal(manager,index)) {
      f.index=index;
      f.address=manager->memory[index].nextFree;
      while(!cuddEqualAdr(f.address,NULL_ADR))
	{
	  f_ptr=cuddPtrPar(manager,f);
	  f_ptr->type.kids.T=central_list;
	  central_list=f;
	  f.address=f_ptr->next;
	}
      manager->memory[index].nextFree=NULL_ADR; 
    }
  manager->ReorderFreeList=central_list;
  return(1);
}

static int cuddInitMemForReconstruct(DdManager *manager,DdMemory **old_mems)
{
  int index;
  DdMemory *mem;
  *old_mems=manager->memory;
  manager->memory=ALLOC(DdMemory,manager->maxSize);
  
  for(index=0;index<manager->maxSize;index++)
    {
      mem=&manager->memory[index];
      mem->TableSize = DD_DEFAULT_PAGES;
      mem->PageTable = ALLOC(DdNodeOld *,DD_DEFAULT_PAGES);
      if(mem->PageTable == NULL)
	return(0);
      mem->nextPage  = 0;
      mem->nextFree  = NULL_ADR;
      mem->allocated = 0;
      manager->allocatedPerLevel[index]= 0;
    }
  manager->memused += manager->maxSize* sizeof(DdNodeOld*) * DD_DEFAULT_PAGES;
  manager->allocated=0;
}

static int cuddRestoreSubtableTypes(DdManager *manager)
{
  int slots,index,i,level;

  for(index=0;index<manager->maxSize;index++)
    {
      level=cuddIndexToLevel(manager,index);
      FREE(manager->subtables[level].ReorderNodelist);
      slots = CUDD_UNIQUE_SLOTS;
      while (slots * DD_MAX_SUBTABLE_DENSITY < manager->subtables[index].keys)
	slots = Cudd_Prime(DD_GROW_FACTOR * slots);

      manager->subtables[level].nodelist=ALLOC(DdNodeAdr,slots);
      if(manager->subtables[level].nodelist==NULL)
	return(0);

      manager->subtables[level].slots=slots;
      for(i=0;i<slots;i++)
	{
	  manager->subtables[level].nodelist[i]=NULL_ADR;
	}
      manager->subtables[level].maxKeys=slots * DD_MAX_SUBTABLE_DENSITY;
    }
  return(1);
}
  
static int cuddRestoreLocalFreeLists(DdManager *manager)
{
  DdNodePar list;
  DdNodeOld *ptr;
  list=manager->ReorderFreeList;
  while(!cuddEqualAdr(list.address,NULL_ADR))
    {
      ptr=cuddPtrPar(manager,list);
      ptr->next=manager->memory[list.index].nextFree;
      manager->memory[list.index].nextFree=list.address;
      list=ptr->type.kids.T;
    }
  return(1);
}

static int cuddFindNewLocations(DdManager *manager,DdMemory *old_mems,
				DdNodeAdr nodelists[])
{
  int index,page_number,number,ext;
  DdNodeOld *page,*f_ptr;
  DdMemory *mem;
  DdNodeMore *more_page;
  DdNodePar f,node;
  int old_nextPage[manager->size];
  
  for(index=0;index<manager->size;index++) {
    if(cuddLocal(manager,index))
      {
	nodelists[index]=NULL_ADR;
      }
  }

  for(index=0;index<manager->size;index++) {
    mem=&old_mems[index];
    if(cuddLocal(manager,index))
      {
	node.index=index;
      	for(page_number=0;page_number<mem->nextPage;page_number++) {
	  page=mem->PageTable[page_number];
	  more_page=mem->PageTable2[page_number];
	  node.address.page=page_number;

	  for(number=1;number<=DD_MEM_CHUNK;number++) {
	    if(more_page[number].Reorder.Ref>0) {
	      ext=more_page[number].Reorder.Ref==DD_EXTERN_REF;
		
	      /* creating the new node at the level corosponding to the
		 index it has ensures that level==index after reorder */
	      f.index=cuddIndexToLevel(manager,more_page[number].Reorder.Index);
	      f.address=cuddAllocNodePar(manager,f.index);
	      f_ptr=cuddPtrPar(manager,f);
	      f_ptr->type.kids.T=page[number].type.kids.T;
	      f_ptr->type.kids.E=page[number].type.kids.E;
	      more_page[number].NewLocation=f;
	      
	      if(ext)
		cuddStoreForSending(manager,f);
	      
	      /* add f to the appropriate list so it can be inserted in 
		 the unique table when the children are known */
	      f_ptr->next=nodelists[f.index];
	      nodelists[f.index]=f.address;
	    }
	  }
	  free(page+1);
	}
      }
    FREE(mem->PageTable);
  }
  return(1);
}


static DdNodePar cuddNewLocation(DdMemory *old_mems,DdNodePar node)
{
  DdNodePar ret;
  if(node.index==CUDD_MAXINDEX)
    return node;

  ret=old_mems[node.index].
    PageTable2[node.address.page][abs(node.address.number)].NewLocation;
 
 if(Cudd_IsComplementPar(node))
    ret=Cudd_NotPar(ret);
  return ret;
}

static int cuddRecreateSubtables(DdManager *manager,DdMemory *old_mems,
				 DdNodeAdr lists[])
{
  DdNodeAdr *nodelist,list;
  int index;
  DdNodeOld *f_ptr;
  DdNodePar f;
  int keys;
  int pos,slots;

  for(index=0;index<manager->size;index++) {
    if(cuddLocal(manager,index))
      {
	f.index=index;
	list=lists[index];
	nodelist=manager->subtables[index].nodelist;
	slots=manager->subtables[index].slots;
	keys=0;
	while(!cuddEqualAdr(list,NULL_ADR))
	  {
	    DdNodePar T,E;
	    f.address=list;
	    f_ptr=cuddPtrPar(manager,f);
	    list=f_ptr->next;
	    T=f_ptr->type.kids.T=cuddNewLocation(old_mems,f_ptr->type.kids.T);
	    E=f_ptr->type.kids.E=cuddNewLocation(old_mems,f_ptr->type.kids.E);

	    pos=ddHashPar(f_ptr->type.kids.T,f_ptr->type.kids.E,slots);
	    f_ptr->next=nodelist[pos];
	    nodelist[pos]=f.address;
	    keys++;
	  }
	assert(keys<=manager->subtables[index].maxKeys);
	manager->subtables[index].keys=keys;
      }
  }
  return(1);
}

static int cuddFreeMore(DdManager *manager,DdMemory *old_mems)
{
  int i,page;
  DdMemory *mem;

  for(i=0;i<manager->size;i++)
    {
      mem=&old_mems[i];
      if(cuddLocal(manager,i))
	{
	  for(page=0;page<mem->nextPage;page++)
	    {
	      free(mem->PageTable2[page]+1);
	      manager->memused-=DD_MEM_CHUNK*sizeof(DdNodeMore);
	    }
	}
      else
	{
	  for(page=0;page<mem->TableSize;page++)
	    if(mem->PageTable2[page]!=NULL)
	      {
		free(mem->PageTable2[page]+1);
		manager->memused-=DD_MEM_CHUNK*sizeof(DdNodeMore);
	      }
	}
      FREE(mem->PageTable2);
      manager->memused-=mem->TableSize*sizeof(DdNodeMore *);
    }
  return(1); /* no allocation, so should not fail */    
}

int cuddDecreaseRefDeath(DdManager *manager,DdNodePar f)
{
  if(cuddDecreaseRef(manager,f)>0)
    return 0;
  else
    {
      cuddDeleteNodePar(manager,f);
      return(1);
    }
}

int swaps=0;

/* swap level with level+1 
   returns the grwoth in nodes
*/
static int cuddSwap(DdManager *manager, int level)
{
  int i;
  int y_index,x_index;
  DdSubtable *subtable;
  DdNodePar *nodelist;
  DdNodePar f;
  DdNodeOld *f_ptr;
  DdNodeAdr *last_adr;
  DdHalfWord *last_index;
  DdNodePar swap_list=CUDD_ERROR;

  swaps++;

  x_index=cuddLevelToIndex(manager,level);
  y_index=cuddLevelToIndex(manager,level+1);

  subtable=&manager->subtables[x_index];
  nodelist=subtable->ReorderNodelist;

  for(i=0;i<subtable->slots;i++)
    {
      last_adr=  &nodelist[i].address;
      last_index=&nodelist[i].index;
      f.address=*last_adr;
      f.index=*last_index;
      while (!cuddEqualAdr(f.address,NULL_ADR)) {
	f_ptr=cuddPtrPar(manager,f);
	if(cuddReorderReadIndex(manager,f_ptr->type.kids.T) == y_index
	   || cuddReorderReadIndex(manager,f_ptr->type.kids.E) == y_index)
	  { 
	    /* remove from table */
	    *last_adr=  f_ptr->next; 
	    *last_index=cuddReorderNextIndex(manager,f);
	    /* add f to swaplist */
	    f_ptr->next=swap_list.address;
	    cuddReorderNextIndex(manager,f)=swap_list.index;
	    swap_list=f;
	    subtable->keys--;
	  }
	else
	  {
	    /* go to next element */
	    last_adr=  &f_ptr->next; /* remove from table */
	    last_index=&cuddReorderNextIndex(manager,f);
	  }
	f.address=*last_adr;
	f.index=*last_index;
      }
    }

  cuddSwapNodes(manager,x_index,y_index,swap_list);

  cuddSetNewLevel(manager,x_index,level+1);
  cuddSetNewLevel(manager,y_index,level);

  return(1);
}

int cuddReorderSplit(DdManager *manager, int index,DdNodePar f,DdNodePar *t,DdNodePar *e)
{
  DdNodeOld *f_ptr;
  int dead;
  if(cuddReorderReadIndex(manager,f)==index)
    {
      f_ptr=cuddPtrPar(manager,f); 
      *t=f_ptr->type.kids.T;
      *e=f_ptr->type.kids.E;
      if(Cudd_IsComplementPar(f))
	{
	  *t=Cudd_NotPar(*t);
	  *e=Cudd_NotPar(*e);
	}
      dead=cuddDecreaseRefDeath(manager,f);
      if(dead)	
	{
	  cuddDecreaseRef(manager,*t);
	  cuddDecreaseRef(manager,*e);
	}
      return TRUE;
    }
  
  cuddDecreaseRef(manager,f);
  *t=*e=f;
  return FALSE;
}

static void cuddSwapNodes(DdManager *manager, int x_index, int y_index,DdNodePar list)
{
  DdNodePar *nodelist;
  DdNodePar f;
  DdNodeOld *f_ptr;
  DdNodePar f1,f0;
  DdNodePar f11,f10,f01,f00;
  int f1_split,f0_split;      
  int pos,slots;
  DdNodePar T,E;

  slots=manager->subtables[y_index].slots;
  nodelist=manager->subtables[y_index].ReorderNodelist;

  while(!cuddEqualAdr(list.address,NULL_ADR))
    {
      f=list;
      f_ptr=cuddPtrPar(manager,f);    
      list.address=f_ptr->next;
      list.index=cuddReorderNextIndex(manager,f);

      f1=f_ptr->type.kids.T;
      f0=f_ptr->type.kids.E;
      
      f1_split=cuddReorderSplit(manager,y_index,f1,&f11,&f10);
      f0_split=cuddReorderSplit(manager,y_index,f0,&f01,&f00);

      assert(f1_split || f0_split); /* else should not appear in list */

      /* note uniqueinter must update references to children */
      T=f_ptr->type.kids.T=cuddUniqueInterSafePar(manager,x_index,f11,f01);
      cuddIncreaseRefSafe(manager,f_ptr->type.kids.T);

      E=f_ptr->type.kids.E=cuddUniqueInterSafePar(manager,x_index,f10,f00);
      cuddIncreaseRefSafe(manager,f_ptr->type.kids.E);
	      
      assert(!cuddEqualPar(f_ptr->type.kids.T,f_ptr->type.kids.E));

      cuddReorderIndex(manager,f)=y_index;

      pos=ddHashPar(f_ptr->type.kids.T,f_ptr->type.kids.E,slots);
      f_ptr->next=nodelist[pos].address;
      cuddReorderNextIndex(manager,f)=nodelist[pos].index;
      nodelist[pos]=f;

      manager->subtables[y_index].keys++;
   }
}

static void cuddSiftUp(DdManager *manager, int index, int top, 
	   int *best_pos, int *best_size)
{
  int level=manager->perm[index];
  
  while(level>top)
    {
      level--;
      cuddSwap(manager,level);
      if(manager->allocated<*best_size) {
	*best_size=manager->allocated;
	*best_pos=level;
      }
      if(manager->allocated>*best_size * DD_MAX_REORDER_GROWTH)
	return;
    }  
}

static void cuddSiftDown(DdManager *manager, int index, int bottom, 
	   int *best_pos, int *best_size)
{
  int level=manager->perm[index];
  
  while(level<bottom)
    {
      cuddSwap(manager,level);
      level++;
      if(manager->allocated<*best_size) {
	*best_size=manager->allocated;
	*best_pos=level;
      }
      if(manager->allocated>*best_size * DD_MAX_REORDER_GROWTH)
	return;
    }  
}


typedef struct pair
{
  int index;
  int profile;
} pair;

int compar(a,b)
pair *a;
pair *b;
{
  return(b->profile - a->profile);
}

static void cuddSiftBlock(DdManager *manager,int block,int first_level, int last_level)
{
  int index;
  int best_pos,test_pos;
  int best_size,test_size;
  pair sift_order[last_level-first_level+1];
  int start_level;
  int i;

  /* note before we start reordering level=index */
  for(index=first_level;index<=last_level;index++)
    {
      sift_order[index-first_level].index=index;
      sift_order[index-first_level].profile=manager->Profile[index];
    }

  qsort(sift_order,last_level-first_level+1,sizeof(pair),compar);

  for(i=0;i<last_level-first_level+1;i++)
    {  
      index=sift_order[i].index;
      best_size=manager->allocated;

      best_pos=start_level=manager->perm[index];

      if(first_level-start_level <= last_level-start_level)
	{
	  cuddSiftUp(manager,index,first_level,&best_pos,&best_size);
	  cuddSiftDown(manager,index,last_level,&best_pos,&best_size);
	  test_pos=best_pos;
	  test_size=best_size;
	  cuddSiftUp(manager,index,best_pos,&test_pos,&test_size);
	}
      else
	{
	  cuddSiftDown(manager,index,last_level,&best_pos,&best_size);
	  cuddSiftUp(manager,index,first_level,&best_pos,&best_size);
	  test_pos=best_pos;
	  test_size=best_size;
	  cuddSiftDown(manager,index,best_pos,&test_pos,&test_size);
	}
      assert(test_pos==best_pos);
      assert(test_size==best_size);
      assert(manager->allocated==best_size);
      printf("old pos %d ",start_level);
      printf("new pos %d ",best_pos); 
      printf("size %d\n",best_size); fflush(stdout);
    }
}

static void cuddBlockSift(DdManager *manager)
{
  int index;
  int first_index,block;

  index=0;
  while(index<manager->size)
    {
      if(cuddLocal(manager,index))
	{
	  block=manager->ReorderBlocks[index];
	  first_index=index;
	  while(index<manager->size 
		&& manager->ReorderBlocks[index]==block)
	    index++;
	  cuddSiftBlock(manager,block,first_index,index-1);
	}
      else
	index++;
    }
  printf("total swaps: %d\n",swaps);
  swaps=0;
}

static int display_status(DdManager *manager, int level)
{
  int index;
  int live=0;
  int forward=0;
  DdNodeMore *ref_page;
  DdNodeOld *page;
  DdMemory *mem;
  int i,j;

  index=cuddLevelToIndex(manager,level);
  mem=&manager->memory[index];
  page=mem->PageTable[0];
  ref_page=mem->PageTable2[0];

  for(i=1;i<=DD_MEM_CHUNK;i++)
    if(ref_page[i].Reorder.Ref>0)
      {
	if(cuddEqualPar(page[i].type.kids.T,CUDD_FORWARD))
	  forward++;
	else
	  {
	    live++;
	    assert(cuddIndexToLevel(manager,cuddReorderIndex(manager,page[i].type.kids.T))>level);
	    assert(cuddIndexToLevel(manager,cuddReorderIndex(manager,page[i].type.kids.E))>level);
	    /* test for dublicates */
	    for(j=1;j<=DD_MEM_CHUNK;j++)
	      if(i!=j && ref_page[j].Reorder.Ref>0)
		{
		  assert(
		   !cuddEqualPar(page[i].type.kids.T,page[j].type.kids.T) ||
		   !cuddEqualPar(page[i].type.kids.E,page[j].type.kids.E));
		}
	  }
      }

  printf("level %d ",level);
  printf("index %d ",index);
  printf("live %d ",live);
  printf("forward %d ",forward);
  printf("keys %d ",manager->subtables[index].keys);
  printf("allocated %d\n",mem->allocated);
}
    
  
  







